import os
from typing import Any, Union

from vrplib.parse import parse_solution


def read_solution(path: Union[str, os.PathLike]) -> dict[str, Any]:
    """
    Reads the solution from the passed-in file path.

    Parameters
    ----------
    path
        The path to the solution file.

    Returns
    -------
    A dictionary that contains the solution data.

    """
    with open(path, "r") as fi:
        return parse_solution(fi.read())



import os
import warnings
from pathlib import Path
from typing import Union
from urllib.request import urlopen

from .constants import CVRPLIB_URL
from .download_utils import find_set, is_vrptw


def download_instance(name: str, path: Union[str, os.PathLike]):
    """
    Downloads an instance file from CVRPLIB and saves it at the specified path.

    Parameters
    ----------
    name
        The name of the instance to download. Should be one of the names
        listed in `vrplib.list_instances()`.
    path
        The path where the instance file should be saved. If a directory is
        specified, the file will be saved in that directory with the original
        file name.
    """
    msg = (
        "The function 'download_instance' is deprecated and will be removed"
        " in the next major version (vrplib v2.0.0)."
    )
    warnings.warn(msg, DeprecationWarning)

    ext = "txt" if is_vrptw(name) else "vrp"
    url = CVRPLIB_URL + f"{find_set(name)}/{name}.{ext}"
    response = urlopen(url, timeout=30)

    instance_text = response.read().decode("utf-8")

    if os.path.isdir(path):
        path = Path(path) / f"{name}.{ext}"

    with open(path, "w") as fi:
        fi.write(instance_text)



import os
from typing import Any, Union

from vrplib.parse import parse_solomon, parse_vrplib


def read_instance(
    path: Union[str, os.PathLike],
    instance_format: str = "vrplib",
    compute_edge_weights: bool = True,
) -> dict[str, Any]:
    """
    Reads the instance from the passed-in file path.

    Parameters
    ----------
    path
        The path to the instance file.
    instance_format
        The instance format, one of ["vrplib", "solomon"]. Default is "vrplib".
    compute_edge_weights
        Whether to calculate edge weights based on instance specifications
        and node coordinates, if not explicitly provided. Defaults to True.

    Returns
    -------
    A dictionary that contains the instance data.
    """
    with open(path, "r") as fi:
        if instance_format == "vrplib":
            return parse_vrplib(fi.read(), compute_edge_weights)
        elif instance_format == "solomon":
            return parse_solomon(fi.read(), compute_edge_weights)

        raise ValueError(f"Format style {instance_format} not known.")



from typing import Union


def text2lines(text: str) -> list[str]:
    """
    Takes a string and returns a list of non-empty, stripped lines. Also
    removes any comment lines from the given string.
    """
    return [
        stripped
        for line in text.splitlines()
        if (stripped := line.strip()) and not stripped.startswith("#")
    ]


def infer_type(s: str) -> Union[int, float, str]:
    try:
        return int(s)
    except ValueError:
        try:
            return float(s)
        except ValueError:
            return s



CVRPLIB_URL = "http://vrp.atd-lab.inf.puc-rio.br/media/com_vrp/instances/"


CVRP_SETS = [
    "A",
    "B",
    "E",
    "F",
    "M",
    "P",
    "CMT",
    "tai",
    "Golden",
    "Li",
    "X",
    "XXL",
]

VRPTW_SETS = [
    "Solomon",
    "HG",
]

XXL_NAMES = ["Antwerp", "Brussels", "Flanders", "Ghent", "Leuven"]
DIMACS_NAMES = ["Loggi", "ORTEC"]



import re
from typing import Union

import numpy as np

from .parse_distances import parse_distances
from .parse_utils import infer_type, text2lines

Instance = dict[str, Union[str, float, np.ndarray]]


def parse_vrplib(text: str, compute_edge_weights: bool = True) -> Instance:
    """
    Parses a VRPLIB instance. An instance consists of two parts:
    1) Specifications: single line of the form <KEY>:<VALUE>.
    2) Data sections: starts with <KEY>_SECTION, followed by rows of values.

    Each specification is parsed as keyword-value pair, where the keyword is
    given lowercase and the value is unmodified. From each data section the
    name is parsed in lower case without the "_SECTION". The data values are
    parsed as numpy arrays, where customer indices are removed (if applicable).

    Parameters
    ----------
    text
        The instance text.
    compute_edge_weights
        Whether to compute edge weights from the node coordinates.
        Defaults to True.

    Returns
    -------
    dict
        The instance data.
    """
    instance = {}

    specs, sections = group_specifications_and_sections(text2lines(text))

    for spec in specs:
        key, value = parse_specification(spec)
        instance[key] = value

    for section in sections:
        section_name, data = parse_section(section, instance)
        instance[section_name] = data  # type: ignore

    if instance and compute_edge_weights and "edge_weight" not in instance:
        # Compute edge weights if there was no explicit edge weight section
        edge_weights = parse_distances([], **instance)  # type: ignore
        instance["edge_weight"] = edge_weights

    return instance


def group_specifications_and_sections(lines: list[str]):
    """
    Groups instance lines into specifications and section parts.
    """
    specs = []
    sections = []

    end_section = 0
    for idx, line in enumerate(lines):
        if "EOF" in line:
            break
        elif idx < end_section:  # Skip all lines of the current section
            continue
        elif ":" in line:
            specs.append(line)
        elif "_SECTION" in line:
            start = lines.index(line)
            end_section = start + 1

            for next_line in lines[start + 1 :]:
                # The current section ends when a next section or an EOF token
                # is found.
                if "_SECTION" in next_line or "EOF" in next_line:
                    break

                end_section += 1

            sections.append(lines[start:end_section])
        else:
            msg = "Instance does not conform to the VRPLIB format."
            raise RuntimeError(msg)

    return specs, sections


def parse_specification(line: str) -> tuple[str, Union[float, str]]:
    """
    Parses a specification line as keyword-value pair, split at the first colon
    occurrence. The keyword is made lowercase and the value is unmodified.
    """
    k, v = [x.strip() for x in re.split("\\s*:\\s*", line, maxsplit=1)]
    return k.lower(), infer_type(v)


def parse_section(
    lines: list, instance: dict
) -> tuple[str, Union[list, np.ndarray]]:
    """
    Parses the data section into numpy arrays.
    """
    section = _remove_suffix(lines[0].strip(), "_SECTION").lower()
    data_ = [[infer_type(n) for n in line.split()] for line in lines[1:]]

    if section == "edge_weight":
        # Parse separately because it may require additional processing
        return section, parse_distances(data_, **instance)  # type: ignore

    if any(len(row) != len(data_[0]) for row in data_):
        # This is a ragged array, so we shortcut to avoid casting to np.array.
        return section, [row[1:] for row in data_]

    data = np.array(data_)

    if section == "depot":
        # Remove -1 end token and renormalize depots to start at zero
        data = data[data != -1] - 1
    else:
        # We remove the customer indices column from non-depot section
        data = data[:, 1:]

    if data.ndim > 1 and data.shape[-1] == 1:
        # Squeeze data sections that contain only one column.
        data = data.squeeze(-1)

    return section, data


def _remove_suffix(name: str, suffix: str):
    return name[: -len(suffix)] if name.endswith(suffix) else name



import os
from typing import TypeVar, Union

import numpy as np

_ArrayLike = TypeVar("_ArrayLike", list, tuple, np.ndarray)


def write_instance(
    path: Union[str, os.PathLike],
    data: dict[str, Union[str, int, float, _ArrayLike]],
):
    """
    Writes a VRP instance to file following the VRPLIB format [1].

    Parameters
    ---------
    path
        The file path.
    data
        A dictionary of keyword-value pairs. For each key-value pair, the
        following rules apply:
        * If ``value`` is a string, integer or float, then it is considered a
          problem specification and formatted as "{keyword}: {value}".
        * If ``value`` is a one or two-dimensional array, then it is considered
          a data section and formatted as
          ```
          {name}
          1 {row_1}
          2 {row_2}
          ...
          n {row_n}
          ```
          where ``name`` is the key and ``row_1``, ``row_2``, etc. are the
          elements of the array. One-dimensional arrays are treated as column
          vectors. If name is "EDGE_WEIGHT_SECTION" or "DEPOT_SECTION", then
          the index is not included.

    References
    ----------
    [1] Helsgaun, K. (2017). An Extension of the Lin-Kernighan-Helsgaun TSP
        Solver for Constrained Traveling Salesman and Vehicle Routing
        Problems.
        http://webhotel4.ruc.dk/~keld/research/LKH-3/LKH-3_REPORT.pdf

    """
    with open(path, "w") as fh:
        for key, value in data.items():
            if isinstance(value, (str, int, float)):
                fh.write(f"{key}: {value}" + "\n")
            else:
                fh.write(_format_section(key, value) + "\n")

        fh.write("EOF\n")


def _format_section(name: str, data: _ArrayLike) -> str:
    """
    Formats a data section.

    Parameters
    ----------
    name
        The name of the section.
    data
        The data to be formatted.

    Returns
    -------
    str
        A VRPLIB-formatted data section.
    """
    section = [name]
    include_idx = name not in ["EDGE_WEIGHT_SECTION", "DEPOT_SECTION"]

    if _is_one_dimensional(data):
        # Treat 1D arrays as column vectors, so each element is a row.
        for idx, elt in enumerate(data, 1):
            prefix = f"{idx}\t" if include_idx else ""
            section.append(prefix + str(elt))
    else:
        for idx, row in enumerate(data, 1):
            prefix = f"{idx}\t" if include_idx else ""
            rest = "\t".join([str(elt) for elt in row])
            section.append(prefix + rest)

    return "\n".join(section)


def _is_one_dimensional(data: _ArrayLike) -> bool:
    for elt in data:
        if isinstance(elt, (list, tuple, np.ndarray)):
            return False
    return True



import os
from typing import Any, Optional, Union


def write_solution(
    path: Union[str, os.PathLike],
    routes: list[list[int]],
    data: Optional[dict[str, Any]] = None,
):
    """
    Writes a VRP solution to file following the VRPLIB convention.

    Parameters
    ----------
    path
        The file path.
    routes
        A list of routes, each route denoting the order in which the customers
        are visited.
    data
        Optional data dictionary. Each key-value pair is written to the
        solution file as "{key}: {value}".
    """
    for route in routes:
        if len(route) == 0:
            raise ValueError("Empty route in solution.")

    with open(path, "w") as fi:
        for idx, route in enumerate(routes, 1):
            text = " ".join([f"Route #{idx}:"] + [str(val) for val in route])
            fi.write(text + "\n")

        if data is not None:
            for key, value in data.items():
                fi.write(f"{key}: {value}\n")



from typing import Union

import numpy as np

from .parse_distances import pairwise_euclidean
from .parse_utils import text2lines

Instance = dict[str, Union[str, float, np.ndarray]]


def parse_solomon(text: str, compute_edge_weights: bool = True) -> Instance:
    """
    Parses the text of a Solomon VRPTW instance.

    Parameters
    ----------
    text
        The instance text.
    compute_edge_weights
        Whether to compute the edge weights from the node coordinates.
        Defaults to True.

    Returns
    -------
    The instance data as dictionary.
    """
    lines = text2lines(text)

    is_valid_solomon_instance(lines)

    instance: Instance = {"name": lines[0]}
    instance["vehicles"], instance["capacity"] = [
        int(num) for num in lines[3].split()
    ]

    data = np.genfromtxt(lines[6:], dtype=int)

    instance["node_coord"] = data[:, 1:3]
    instance["demand"] = data[:, 3]
    instance["time_window"] = data[:, 4:6]
    instance["service_time"] = data[:, 6]

    if compute_edge_weights:
        instance["edge_weight"] = pairwise_euclidean(instance["node_coord"])

    return instance


def is_valid_solomon_instance(lines: list[str]):
    """
    Checks if the passed-in lines follow the Solomon format requirements.
    """

    try:
        assert lines[0]  # non-empty first line
        assert "VEHICLE" in lines[1]
        assert "NUMBER" in lines[2]
        assert "CAPACITY" in lines[2]
        assert "CUSTOMER" in lines[4]

        # Header names are separated on whitespace because the spacing of
        # some Solomon instances is off.
        headers = [
            "CUST",
            "NO.",
            "XCOORD.",
            "YCOORD.",
            "DEMAND",
            "READY",
            "DUE",
            "DATE",
            "SERVICE",
            "TIME",
        ]
        for header in headers:
            assert header in lines[5]

    except (IndexError, ValueError, AssertionError) as err:
        msg = "Instance does not conform to the Solomon format."
        raise RuntimeError(msg) from err



import importlib.resources as pkg_resource
import warnings
from functools import lru_cache
from typing import Optional

from .download_utils import is_vrptw


def list_names(
    low: Optional[int] = None,
    high: Optional[int] = None,
    vrp_type: Optional[str] = None,
):
    """
    Returns the names of the instances that can be downloaded from CVRPLIB.

    Parameters
    ----------
    low
        The minimum number of customers of the listed instances.
    high
        The maximum number of customers of the listed instances.
    vrp_type
        The vrp_type, one of ['cvrp', 'vrptw']. If not set, then instances
        of both types are returned.
    """
    instances = _read_instance_data()
    msg = (
        "The function 'list_names' is deprecated and will be removed"
        " in the next major version (vrplib v2.0.0)."
    )
    warnings.warn(msg, DeprecationWarning)

    if low is not None:
        instances = filter(lambda inst: inst["n_customers"] >= low, instances)

    if high is not None:
        instances = filter(lambda inst: inst["n_customers"] <= high, instances)

    if vrp_type not in [None, "cvrp", "vrptw"]:
        raise ValueError("vrp_type must be one of [None, 'cvrp', 'vrptw']")

    elif vrp_type == "cvrp":
        instances = filter(lambda inst: not is_vrptw(inst["name"]), instances)

    elif vrp_type == "vrptw":
        instances = filter(lambda inst: is_vrptw(inst["name"]), instances)

    return [inst["name"] for inst in instances]


@lru_cache()
def _read_instance_data():
    """
    Reads the instance data. All CVRPLIB instance names are stored in the
    `instance_data.csv` file.
    """
    fi = pkg_resource.read_text(__package__, "instance_data.csv")
    instances = [line.strip().split(",") for line in fi.split()]

    return [
        {"name": inst[0], "n_customers": int(inst[1])} for inst in instances
    ]



from itertools import combinations
from typing import Optional, Union

import numpy as np


def parse_distances(
    data: list[float],
    edge_weight_type: str,
    edge_weight_format: Optional[str] = None,
    node_coord: Optional[np.ndarray] = None,
    comment: Optional[str] = None,
    **kwargs: Union[float, str, np.ndarray],  # noqa
) -> np.ndarray:
    """
    Parses the distances. The specification "edge_weight_type" describes how
    the distances should be parsed. The two main ways are to calculate the
    Euclidean distances using the the node coordinates or by parsing an
    explicit distance matrix.

    Parameters
    ----------
    data
        The edge weight data.
    edge_weight_type
        The type of the edge weight data.
    edge_weight_format, optional
        The format of the edge weight data.
    node_coord, optional
        The customer location coordinates.
    comment, optional
        The comment specification in the instance.
    **kwargs, optional
        Optional keyword arguments.

    Returns
    -------
    np.ndarray
        An n-by-n distances matrix.
    """
    if "2D" in edge_weight_type:  # Euclidean distance on node coordinates
        if node_coord is None:
            msg = (
                "Cannot compute Euclidean distances because node coordinates "
                "are not provided."
            )
            raise ValueError(msg)

        distance = pairwise_euclidean(node_coord)

        if edge_weight_type == "EUC_2D":
            return distance

        if edge_weight_type == "FLOOR_2D":
            return np.floor(distance)

        if edge_weight_type == "EXACT_2D":
            return np.round(distance * 1000)

    if edge_weight_type == "EXPLICIT":
        if edge_weight_format == "LOWER_ROW":
            # TODO Eilon instances edge weight specifications are incorrect in
            # (C)VRPLIB format. Find a better way to identify Eilon instances.
            if comment is not None and "Eilon" in comment:
                return from_eilon(data)
            else:
                return from_lower_row(data)

        if edge_weight_format == "FULL_MATRIX":
            return np.array(data)

    raise ValueError("Edge weight type or format unknown.")


def pairwise_euclidean(coords: np.ndarray) -> np.ndarray:
    """
    Computes the pairwise Euclidean distance between the passed-in coordinates.

    Parameters
    ----------
    coords
        An n-by-2 array of location coordinates.

    Returns
    -------
    np.ndarray
        An n-by-n Euclidean distances matrix.

    """
    diff = coords[:, np.newaxis, :] - coords
    square_diff = diff**2
    square_dist = np.sum(square_diff, axis=-1)
    return np.sqrt(square_dist)


def from_lower_row(triangular: np.ndarray) -> np.ndarray:
    """
    Computes a full distances matrix from a lower row triangular matrix.
    The triangular matrix should not contain the diagonal.

    Parameters
    ----------
    triangular
        A list of lists, each list representing the entries of a row in a
        lower triangular matrix without diagonal entries.

    Returns
    -------
    np.ndarray
        A n-by-n distances matrix.
    """
    n = len(triangular) + 1
    distances = np.zeros((n, n))

    for i in range(n - 1):
        distances[i + 1, : i + 1] = triangular[i]

    return distances + distances.T


def from_eilon(edge_weights: np.ndarray) -> np.ndarray:
    """
    Computes a full distances matrix from the Eilon instances with "LOWER_ROW"
    edge weight format. The specification is incorrect, instead the edge weight
    section needs to be parsed as a flattend, column-wise triangular matrix.

    See https://github.com/leonlan/VRPLIB/issues/40.
    """
    flattened = [dist for row in edge_weights for dist in row]
    n = int((2 * len(flattened)) ** 0.5) + 1  # The (n+1)-th triangular number

    distances = np.zeros((n, n))
    indices = sorted([(i, j) for (i, j) in combinations(range(n), r=2)])

    for idx, (i, j) in enumerate(indices):
        d_ij = flattened[idx]
        distances[i, j] = d_ij
        distances[j, i] = d_ij

    return distances


def is_triangular_number(n):
    """
    Checks if n is a triangular number.
    """
    i = int((2 * n) ** 0.5)
    return i * (i + 1) == 2 * n



import re

from .constants import CVRP_SETS, DIMACS_NAMES, XXL_NAMES


def find_set(instance_name: str) -> str:
    """
    Find the set name of the instance.

    Notes
    -----
    - VRPTW instances start with "C, R, RC" directly followed by 1 or 2.
        HG instances have underscores ("_") in the name, whereas Solomon
        instances do not.
    - CVRP instance names and their corresponding set names share the same
        first letter. the exceptions are XXL and DIMACS instances, which have
        unique instance names
    """
    if re.match("(R|C|RC)[12]", instance_name):
        if "_" in instance_name:
            return "HG"
        else:
            return "Solomon"

    if any([instance_name.startswith(xxl) for xxl in XXL_NAMES]):
        return "XXL"

    if any([instance_name.startswith(dimacs) for dimacs in DIMACS_NAMES]):
        return "D"

    for set_name in CVRP_SETS:
        if instance_name.startswith(set_name):
            return set_name

    raise ValueError(f"Set name not known for instance: {instance_name}.")


def is_vrptw(name: str) -> bool:
    """
    Checks if the passed-in name is a VRPTW instance or not. Otherwise the
    instance is a CVRP instance.
    """
    return find_set(name) in ["HG", "Solomon"]



from typing import Union

from .parse_utils import infer_type, text2lines

Solution = dict[str, Union[float, str, list]]


def parse_solution(text: str) -> Solution:
    """
    Parses the text of a solution file formatted in VRPLIB style. A solution
    consists of routes, which are indexed from 1 to n, and possibly other data.

    Parameters
    ----------
    text
        The solution text.

    Returns
    -------
    dict
        The soluion data.
    """
    solution: Solution = {"routes": []}

    for line in text2lines(text):
        if "Route" in line:
            route = [int(idx) for idx in line.split(":")[1].split(" ") if idx]
            solution["routes"].append(route)  # type: ignore
        elif ":" in line or " " in line:  # Split at first colon or whitespace
            split_at = ":" if ":" in line else " "
            k, v = [word.strip() for word in line.split(split_at, 1)]
            solution[k.lower()] = infer_type(v)
        else:  # Ignore lines without keyword-value pairs
            continue

    return solution
